from model import *
from utils import *
from dataset import *
from torch.utils.data import DataLoader
import pandas as pd
from tqdm import tqdm
from evaluate import load
import wandb
from config import *

device="cuda:2"

if __name__ == '__main__':
    #define model
    torch.set_default_dtype(torch.float32)
    torch.autograd.set_detect_anomaly(True)
    # model = VNTrOCR()
    # model.load_state_dict(torch.load("VisionEncoderDecoder/weights/cp_finetune_v2_5.pt", map_location=device))
    model = AdapterVNTrOCR(config)
    model.load_state_dict(torch.load("VisionEncoderDecoder/weights/cp_add_finetune_v2.pt", map_location=device))
    model.to(device)
    
    valdataset = CustomDataset(pd.read_csv("test_anot.csv"), config)
    valloader = DataLoader(dataset=valdataset, batch_size=128, shuffle=False, generator=torch.Generator(device="cpu"), num_workers=8)


    #loss function
    # lsLoss = LabelSmoothingLoss(233, padding_idx=0, smoothing=0.1)

    #utils
    cer = load("cer")
    wer = load("wer")
    cer_score = 0
    wer_score = 0

    # wandb init
    # wandb.login(key = '9d97bfb2299059fecbd3e48dd349fafce5906d76')
    # wandb.login(key = 'e0e0a2547f255a36f551d7b6a166b84e5139d276')
    # wandb.init(
    #   project="fine-tune-VNHTR",
    #   name=f"vision_encoder_decoder_augmented_data",
    #   config={
    #       "learning_rate": 0.0001,
    #       "architecture": "v_t",
    #       "dataset": "gen",
    #       "epochs": 10,
    #   }
    # )


    print("Start .....")
    with torch.no_grad():
        model.eval()
        predictions, references = [], []
        # torch.cuda.empty_cache()
        for batch in tqdm(valloader):
            pixel_values = batch["pixel_values"].to(device)
            input_ids = batch["input_ids"].to(device)

            preds = infer(pixel_values, model, adapter=True)

            predictions += valdataset.tokenizer.batch_decode(preds, skip_special_tokens=True)
            references += valdataset.tokenizer.batch_decode(input_ids, skip_special_tokens=True)
        cer_score += cer.compute(predictions=predictions, references=references, concatenate_texts=True)
        wer_score += wer.compute(predictions=predictions, references=references, concatenate_texts=True)
        print(f'Character Error Rate: {cer_score}')
        print(f'Word Error Rate: {wer_score}')


from dataclasses import dataclass
from vietocr.tool.predictor import Predictor
from vietocr.tool.config import Cfg
import torch

# device = torch.device("cuda")
device="cpu"

@dataclass
class Config:
    n_decoder_blocks: int = 6
    n_encoder_blocks: int = 12
    im_height: int = 384
    im_width: int = 384
    n_embd: int = 768
    max_seq_len: int = 35
    in_chanel: int = 3
    rank: int = 16
    vocab_size: int = 40030
    bias: bool = False

config = Config()



from jit import TrOCR
import torch
from PIL import Image
from transformers import AutoTokenizer, TrOCRProcessor
import numpy as np

device = "cuda:2"

if __name__ == '__main__':
    predictor = TrOCR(device)
    torch.cuda.set_device(3)
    img = Image.open("/mnt/disk4/VN_HTR/VN_HTR/test/test11_test1.jpg")
    processor = TrOCRProcessor.from_pretrained("microsoft/trocr-small-handwritten")
    dummy_input = processor(img, return_tensors="pt").pixel_values.to(device)

    # INIT LOGGERS
    starter, ender = torch.cuda.Event(enable_timing=True, ), torch.cuda.Event(enable_timing=True)
    repetitions = 300
    timings=np.zeros((repetitions,1))
    #GPU-WARM-UP
    print("warming up ...")
    for _ in range(10):
        _ = predictor.predict(dummy_input)
    # MEASURE PERFORMANCE
    print("measuring ...")
    with torch.no_grad():
        for rep in range(repetitions):
            starter.record()
            _ = predictor.predict(dummy_input)
            ender.record()
            # WAIT FOR GPU SYNC
            torch.cuda.synchronize()
            curr_time = starter.elapsed_time(ender)
            timings[rep] = curr_time

    mean_syn = np.sum(timings) / repetitions
    std_syn = np.std(timings)
    print(mean_syn)


from jit import OCRModel
import torch
from PIL import Image
from dataset import process_input
import numpy as np

device = "cuda:1"

if __name__ == '__main__':
    predictor = OCRModel(device)
    torch.cuda.set_device(3)

    img = Image.open("/mnt/disk4/VN_HTR/VN_HTR/test/test11_test1.jpg")
    img, _ = process_input(img)

    dummy_input = torch.FloatTensor(img).unsqueeze(0).to(device)

    # INIT LOGGERS
    starter, ender = torch.cuda.Event(enable_timing=True, ), torch.cuda.Event(enable_timing=True)
    repetitions = 300
    timings=np.zeros((repetitions,1))
    #GPU-WARM-UP
    print("warming up ...")
    for _ in range(10):
        _ = predictor.predict(dummy_input)
    # MEASURE PERFORMANCE
    print("measuring ...")
    for rep in range(repetitions):
        starter.record()
        _ = predictor.predict(dummy_input)
        ender.record()
        # WAIT FOR GPU SYNC
        torch.cuda.synchronize()
        curr_time = starter.elapsed_time(ender)
        timings[rep] = curr_time

    mean_syn = np.sum(timings) / repetitions
    std_syn = np.std(timings)
    print(mean_syn)


import os
import random
import numpy as np
from tqdm import  tqdm


def rand_parts(seq, n):
    indices = range(len(seq) - (8 - 1) * n)
    result = []
    offset = 0
    ls = np.random.poisson(6, n)
    j = 0
    for i in tqdm(sorted(random.sample(indices, n))):
        l = max(2, min(ls[j], 12))
        j += 1
        i += offset
        try:
            result.append(" ".join(seq[i:i+l]))
        except:
            result.append(" ".join(seq[i:]))
            break
        offset += l - 1
    return result

def process_folder(path):
    fns = os.listdir(path)
    res = []
    for fn in tqdm(fns):
        f = open(path + "/" + fn, 'r', encoding='utf-16')
        content = f.read()
        f.close()
        ignore_char = ['!', '"', '%', '&', '(', ')', '*', ',', '-', '.', '/', ':', ';', '?']
        for char in ignore_char:
            content = content.replace(char, ' ')
        sentences = content.split()
        res += rand_parts(sentences, round(len(sentences)/10))

    with open("sentences.txt", 'a', encoding='utf-16') as f:
        f.write("\n".join(res) + "\n")
        f.close()

if __name__ == "__main__":
    for folder in os.listdir("VNTC"):
        process_folder("VNTC/" + folder)

    # res = []
    # f = open("corpus-title.txt", 'r', encoding='utf-8')
    # content = f.read()
    # f.close()
    # ignore_char = ['!', '"', '%', '&', '(', ')', '*', ',', '-', '.', '/', ':', ';', '?']
    # for char in ignore_char:
    #     content = content.replace(char, ' ')
    # sentences = content.split()
    # try:
    #     res += rand_parts(sentences, round(len(sentences) / 8))
    # except:
    #     pass
    # with open("sentences.txt", 'a', encoding='utf-16') as f:
    #     f.write("\n".join(res) + "\n")
    #     f.close()



# from model import *
# from utils import *
# from dataset import *
# from torch.utils.data import DataLoader
# import pandas as pd
# from tqdm import tqdm
# from evaluate import load
# from config import config
import numpy as np
import torch.nn as nn
from torch.nn import functional as F
import torch

class TrOCR():
    def __init__(self, adapter=False, device="cuda:2"):
        if adapter:
            self.encoder = torch.jit.load("vnhtr/weights/tra_encoder.pt", map_location=device)
            self.decoder = torch.jit.load("vnhtr/weights/tra_decoder.pt", map_location=device)
        else:
            self.encoder = torch.jit.load("vnhtr/weights/tr_encoder.pt", map_location=device)
            self.decoder = torch.jit.load("vnhtr/weights/tr_decoder.pt", map_location=device)
        self.encoder.eval()
        self.decoder.eval()
        self.device = device
    
    def predict(self, pixel_values, max_seq_length=35, sos_token=0, eos_token=2):
        with torch.no_grad():
            with torch.jit.optimized_execution(True):
                encoder_output = self.encoder(pixel_values.to(self.device))
    
                start_ids = torch.LongTensor([[sos_token]]*pixel_values.shape[0]).to(self.device)

                max_length = 0

                while max_length <= max_seq_length and not all(start_ids[:, -1] == eos_token):
                    output = self.decoder(start_ids, encoder_output, self.gen_nopeek_mask(start_ids.shape[1]).to(self.device))
                    output = F.softmax(output[:,-1,:], dim=-1)

                    start_ids = torch.cat([start_ids, output.argmax(dim=-1).unsqueeze(1)], dim=-1)

                    max_length += 1
            
            return start_ids
    
    def predict_topk(self, pixel_values, max_seq_length=35, sos_token=0, eos_token=2):
        "data: BxCXHxW"
        conflict = []
        with torch.no_grad():
            encoder_output = self.encoder(pixel_values.to(self.device))
    
            start_ids = torch.LongTensor([[sos_token]]*pixel_values.shape[0]).to(self.device)

            max_length = 0

            probs = []
            tokens = []

            while max_length <= max_seq_length and not all(start_ids[:, -1] == eos_token):
                output = self.decoder(start_ids, encoder_output, self.gen_nopeek_mask(start_ids.shape[1]).to(self.device))
                output = F.softmax(output[:,-1,:], dim=-1)

                start_ids = torch.cat([start_ids, output.argmax(dim=-1).unsqueeze(1)], dim=-1)

                _prob, _token = output.topk(k=501, dim=-1)
                # print(_prob.shape)
                probs.append(_prob[0].cpu().numpy())
                tokens.append(_token[0].cpu().numpy())

                max_length += 1

                # conflict.append(output[:,].cpu().numpy())
        
        return start_ids, np.array(tokens), np.array(probs), np.array(probs)
    
    def gen_nopeek_mask(self, length):
        mask = (torch.triu(torch.ones(length, length)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask

class ViT(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.vit = model.basemodel.ViTEncoder
        self.adaptive_layer = model.basemodel.adaptive_layer

    def forward(self, pixel_values):
        encoder_output = self.vit(pixel_values)[0] # B, src_len, 384
        encoder_output = self.adaptive_layer(encoder_output) # B, src_len 768
        return encoder_output

class TrocrDecoder(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.BartDecoder = model.basemodel.BartDecoder
        self.fc = model.basemodel.fc_out
        self.rethinking = model.rethinking
        # self.BartDecoder = model.BartDecoder
        # self.fc = model.fc_out

    def forward(self, prev_ids, encoder_output, mask):
        decoder_output = self.BartDecoder(input_ids=prev_ids,
                                          encoder_hidden_states=encoder_output) # B, tgt_len, 768
        logits = self.fc(decoder_output[0]) # B, tgt_len, vocab_size
        logits = logits + self.rethinking(logits, mask)
        return logits
    
def gen_nopeek_mask(length):
    mask = (torch.triu(torch.ones(length, length)) == 1).transpose(0, 1)
    mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
    return mask

device = 'cuda:0'

if __name__ == "__main__":
    # model = VNTrOCR()
    # model.load_state_dict(torch.load("/mnt/disk4/VN_HTR/VN_HTR/VisionEncoderDecoder/weights/cp_finetune_v2_5.pt", map_location=device))
    model = AdapterVNTrOCR(config)
    model.load_state_dict(torch.load("/mnt/disk4/VN_HTR/VN_HTR/VisionEncoderDecoder/weights/cp_add_finetune_v2.pt", map_location=device))
    model.to(device)

    anot = pd.read_csv("test_anot.csv").sample(frac=1, random_state=0).reset_index(drop=True)

    valdataset = CustomDataset(anot, config)
    valloader = DataLoader(dataset=valdataset, batch_size=2, shuffle=False, generator=torch.Generator(device="cpu"))

    for batch in valloader:
        pixel_values = batch["pixel_values"].to(device)
        input_ids = batch["input_ids"].to(device)

        encoder = ViT(model)
        encoder.eval()
        trace_encoder = torch.jit.trace(encoder, pixel_values)
        trace_encoder.save("/mnt/disk4/VN_HTR/VN_HTR/vnhtr/weights/tra_encoder.pt")
        memory = encoder(pixel_values)#.transpose(1, 0)
        # print(memory.shape)
        # print(in_tokens[:, :5].shape)

        decoder = TrocrDecoder(model)
        decoder.eval()
        trace_decoder = torch.jit.trace(decoder, (input_ids[:,:2], memory, gen_nopeek_mask(input_ids[:,:2].shape[1]).to(device)))
        trace_decoder.save("/mnt/disk4/VN_HTR/VN_HTR/vnhtr/weights/tra_decoder.pt")
        
        break


import torch
import numpy as np
from torch.nn import functional as F

def infer_transformer_base(img, net, max_seq_length=128, sos_token=1, eos_token=2, backbone="vgg", adapter=False):
    "data: BxCXHxW"
    net.eval()
    device = img.device
    net.to(device)
    if not adapter:
        model = net
    else:
        model = net.basemodel

    with torch.no_grad():
        src = model.backbone(img)
        if backbone != "vgg":
            src = src.transpose(1,0)
        
        memory = model.transformerLM.forward_encoder(src)
        
    
        translated_sentence = [[sos_token]*len(img)]
        char_probs = [[1]*len(img)]

        max_length = 0

        while max_length <= max_seq_length and not all(np.any(np.asarray(translated_sentence).T==eos_token, axis=1)):

            tgt_inp = torch.LongTensor(translated_sentence).to(device)
            
            output, memory = model.transformerLM.forward_decoder(tgt_inp, memory)
            if adapter:
                output = net.forward_adapt(output)
            output = F.softmax(output, dim=-1)
            output = output.to('cpu')

            values, indices  = torch.topk(output, 5)
            
            indices = indices[:, -1, 0]
            indices = indices.tolist()
            
            values = values[:, -1, 0]
            values = values.tolist()
            char_probs.append(values)

            translated_sentence.append(indices)   
            max_length += 1

            del output

        translated_sentence = np.asarray(translated_sentence).T
    
    return translated_sentence

def param_to_update(net):
        encoder_param = []
        decoder_param = []
        for name, param in net.named_parameters():
            param = param.float()
            param.requires_grad = True
            if "backbone" in name:
                encoder_param.append(param)
            else:
                decoder_param.append(param)
        return encoder_param, decoder_param


import pandas as pd
import os

# df = pd.read_csv("augmented_anot.csv").sample(frac=1, random_state=42).reset_index(drop=True)
# df1 = pd.read_csv("augmented_anot.csv").sample(frac=1, random_state=42).reset_index(drop=True)[:50000]
df2 = pd.read_csv("wild_anot.csv")
df3 = pd.read_csv("d1_anot.csv")
df4 = pd.read_csv("digits.csv")

df = pd.concat([df2, df3, df4], ignore_index=True).sample(frac=1, random_state=42).reset_index(drop=True)

# _df = pd.concat([df1, df[:-3000]], ignore_index=True).sample(frac=1, random_state=42).reset_index(drop=True)
# print(len(_df))
# _df.to_csv("finetune_anot.csv", index=False)
df[-3000:].to_csv("test_anot.csv", index=False)

# print(len(os.listdir("WildLine")))

# df3 = pd.read_csv("real_anot.csv")
# df4 = pd.read_csv("single_digit.csv")[:20000]

# _df = pd.concat([df3, df4], ignore_index=True).sample(frac=1, random_state=42).reset_index(drop=True)
# print(len(_df))
# _df.to_csv("real_anot.csv", index=False)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="vnhtr",
    version="0.1.4",
    author="nguyenhoanganh2002",
    author_email="anh.nh204511@gmail.com",
    description="Encoder-Decoder base for Vietnamese handwriting recognition",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nguyenhoanganh2002/vnhtr",
    packages=setuptools.find_packages(),
    install_requires=[
        'pillow',
        'transformers==4.35.0',
        'sentencepiece==0.1.99'
    ],
    keywords=['ocr', 'vnocr', 'htr', 'vnhtr'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import json
import cv2
import numpy as np
import random
from tqdm import tqdm
import pandas as pd

with open("word_anot.json", encoding='utf-8') as fh:
    data = json.load(fh)

vocab = {}
for word in tqdm(data.keys()):
    fns = data[word]
    w_im = [cv2.imread("grey_word/w_" + fn).astype(np.int32) for fn in fns]
    vocab[word] = w_im

def resize_n_pad(im):
    im = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)

    # d_h_size = 64
    # max_w = 1024
    #
    # old_size = im.shape[:2]  # old_size is in (height, width) format
    #
    # ratio = 1.0 * d_h_size / old_size[0]
    # new_size = [int(x * ratio) for x in old_size]
    # new_size[1] = min(max_w, new_size[1])
    # new_size[0] = 64
    #
    # # new_size should be in (width, height) format
    # im = cv2.resize(im, (new_size[1], new_size[0]))
    #
    # delta_w = max_w - new_size[1]
    #
    # new_im = cv2.copyMakeBorder(im, 0, 0, 0, delta_w, cv2.BORDER_CONSTANT, value=255)
    return im.astype(np.uint8)

def gen_image(sub_sentence):
    words = sub_sentence.split()
    ims = []
    label = []
    for word in words:
        try:
            image = random.choice(vocab[word])
        except:
            continue
        if image.shape[0] == 63:
            image = np.concatenate((image, image[0:1,:,:]), axis=0)
        image += (random.choice(np.arange(140,171))-np.mean(image).astype(np.int32))
        ims.append(image)
        label.append(word)
    # cv2.imshow("test", np.concatenate(ims, axis=1).astype(np.uint8))
    # cv2.waitKey()
    im = np.concatenate(ims, axis=1).astype(np.uint8)
    im = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
    return im.astype(np.uint8), " ".join(label)

def get_content(sub_sentence):
    words = sub_sentence.split()
    contents = []
    for word in words:
        try:
            image = random.choice(vocab[word])
            contents.append(word)
        except:
            continue
    return " ".join(contents)


if __name__ == "__main__":
    f = open("sentences.txt", 'r', encoding='utf-16')
    contents = f.read().split("\n")
    f.close()
    anot_json = {"filename": [], "label": []}
    for i, content in tqdm(enumerate(contents[:5000000])):
        try:
            new_image, label = gen_image(content)
        except:
            continue
        if len(label) != 0:
            anot_json["filename"].append(f"gen_{i}.jpg")
            anot_json["label"].append(label)
            cv2.imwrite(f"images_gen/gen_{i}.jpg", new_image)
        if i % 100000 == 0:
            pd.DataFrame(data=anot_json).to_csv("gen_anot.csv", index=False)
    pd.DataFrame(data=anot_json).to_csv("gen_anot.csv", index=False)


import torch.nn as nn
import torch
from config import predictor, device
from submodules import *

class AdapterVGGTransformer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.basemodel = VGGTransformer(config)

        self.rethinking = Rethinking(config)
        self.fc_out = nn.Linear(config.vocab_size, config.vocab_size)
    
    def forward(self, img, tgt, tgt_key_padding_mask):
        en_out = self.basemodel.backbone(img)
        de_out = self.basemodel.transformerLM(en_out, tgt.transpose(1,0), tgt_key_padding_mask=tgt_key_padding_mask)
        # print("------------here 4------------")
        de_out = de_out + self.rethinking(de_out)
        return self.fc_out(de_out)
    
    def forward_adapt(self, de_out):
        de_out = de_out + self.rethinking(de_out)
        return self.fc_out(de_out)
        # tgt_mask = self.basemodel.transformerLM.gen_nopeek_mask(tgt.shape[0]).to(img.device)


class VGGTransformer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.max_seq_len = config.max_seq_len
        self.backbone = predictor.cnn.float()

        self.transformerLM = TransformerLM(config)
    
    def forward(self, img, tgt, tgt_key_padding_mask):
        en_out = self.backbone(img)
        return self.transformerLM(en_out, tgt.transpose(1,0), tgt_key_padding_mask=tgt_key_padding_mask)


class ResnetTransformer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.max_seq_len = config.max_seq_len
        self.backbone = Resnet34(config)

        self.transformerLM = TransformerLM(config)
        
        self.softmax = nn.Softmax(dim=-1)
    
    def forward(self, img, tgt, tgt_key_padding_mask):
        en_out = self.backbone(img)
        return self.transformerLM(en_out.transpose(1,0), tgt.transpose(1,0), tgt_key_padding_mask=tgt_key_padding_mask)

class GCTransformer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.max_seq_len = config.max_seq_len
        self.backbone = ConvEmbeddingGC()

        self.transformerLM = TransformerLM(config)
        
        self.softmax = nn.Softmax(dim=-1)
    
    def forward(self, img, tgt, tgt_key_padding_mask):
        en_out = self.backbone(img)
        return self.transformerLM(en_out.transpose(1,0), tgt.transpose(1,0), tgt_key_padding_mask=tgt_key_padding_mask)


class FCNTransformer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.max_seq_len = config.max_seq_len

        self.backbone = FCN_Encoder({"dropout": 0.1, "input_channels": 3})

        self.backbone.load_state_dict(torch.load("source/weights/iam_line.pt", map_location=device)["encoder_state_dict"])

        self.transformerLM = TransformerLM(config)
        
        self.softmax = nn.Softmax(dim=-1)
    
    def forward(self, img, tgt, tgt_key_padding_mask):
        en_out = self.backbone(img)
        return self.transformerLM(en_out.transpose(1,0), tgt.transpose(1,0), tgt_key_padding_mask=tgt_key_padding_mask)